@@ -8,12 +8,11 @@ import operator
 import os
 import re
 import sys
-import typing
-from dataclasses import asdict, dataclass, replace
+from dataclasses import astuple, dataclass, replace
 from functools import lru_cache, partial, reduce
 from typing import Any, Callable, Dict, Generic, Optional, Type, TypeVar
 
-from pytypeclass import Monad, MonadPlus
+from pytypeclass import Monad, MonadPlus, Monoid
 from pytypeclass.nonempty_list import NonemptyList
 
 from dollar_lambda.error import (
@@ -24,11 +23,12 @@ from dollar_lambda.error import (
     UnequalError,
     UnexpectedError,
 )
-from dollar_lambda.key_value import KeyValue, KeyValueTuple
 from dollar_lambda.result import Result
-from dollar_lambda.sequence import Sequence
+from dollar_lambda.sequence import KeyValue, Output, Sequence
 
 A_co = TypeVar("A_co", covariant=True)
+A_monoid = TypeVar("A_monoid", bound=Monoid)
+B_monoid = TypeVar("B_monoid", bound=Monoid)
 A = TypeVar("A")
 B = TypeVar("B")
 
@@ -54,18 +54,9 @@ class Parse(Generic[A_co]):
 
 
 @dataclass
-class SuccessError(ArgumentError, Generic[A]):
+class SuccessError(ArgumentError, Generic[A_monoid]):
     input: Sequence[str]
-    output: NonemptyList[Parse[A]]
-
-
-def empty() -> "Parser[Sequence]":
-    """
-    Always returns {}, no matter the input. Mostly useful for use in `nonpositional`.
-    >>> empty().parse_args("any", "arguments", allow_unparsed=True)
-    {}
-    """
-    return Parser[Sequence[A]].empty()
+    output: NonemptyList[Parse[A_monoid]]
 
 
 def binary_usage(a: Optional[str], op: str, b: Optional[str], add_brackets=True):
@@ -90,8 +81,8 @@ class Parser(MonadPlus[A_co]):
     helps: Dict[str, str]
 
     def __add__(
-        self: "Parser[Sequence[A]]", other: "Parser[Sequence[B]]"
-    ) -> "Parser[Sequence[A | B]]":
+        self: "Parser[Output[A_monoid]]", other: "Parser[Output[B_monoid]]"
+    ) -> "Parser[Output[A_monoid | B_monoid]]":
         """
         Parse two arguments in either order.
         >>> p = flag("verbose") + flag("debug")
@@ -119,28 +110,28 @@ class Parser(MonadPlus[A_co]):
         >>> p.parse_args("-a", "-c", "-b")
         {'a': True, 'c': True, 'b': True}
         """
-        p = (self >> other) | (other >> self)
+        p = (self >> other) | (other.rshift(self))
         usage = binary_usage(self.usage, " ", other.usage, add_brackets=False)
         return replace(p, usage=usage)
 
-    def __ge__(self, f: Callable[[A_co], Monad[B]]) -> "Parser[B]":
+    def __ge__(self, f: Callable[[A_co], Monad[B_monoid]]) -> "Parser[B_monoid]":  # type: ignore[override]
         return self.bind(f)
 
     def __or__(  # type: ignore[override]
-        self: "Parser[A_co]",
-        other: "Parser[B]",
-    ) -> "Parser[A_co | B]":
+        self: "Parser[A_monoid]",
+        other: "Parser[B_monoid]",
+    ) -> "Parser[A_monoid | B_monoid]":
         """
         Tries apply the first parser. If it fails, tries the second. If that fails, the parser fails.
 
-        >>> from dollar_lambda import argument, option, done, flag
+        >>> from dollar_lambda import argument, option, flag
         >>> p = option("option") | flag("verbose")
         >>> p.parse_args("--option", "x")
         {'option': 'x'}
         >>> p.parse_args("--verbose")
         {'verbose': True}
 
-        Note that by default, `parse_args` adds `>> done` to the end of parsers, causing
+        Note that by default, `parse_args` adds `>> Parser.done()` to the end of parsers, causing
         `p` to fail when both arguments are supplied:
 
         >>> p.parse_args("--verbose", "--option", "x")
@@ -152,7 +143,7 @@ class Parser(MonadPlus[A_co]):
         {'verbose': True}
         """
 
-        def f(cs: Sequence[str]) -> Result[Parse["A_co | B"]]:
+        def f(cs: Sequence[str]) -> Result[Parse["A_monoid | B_monoid"]]:
             return self.parse(cs) | other.parse(cs)
 
         return Parser(
@@ -161,9 +152,14 @@ class Parser(MonadPlus[A_co]):
             helps={**self.helps, **other.helps},
         )
 
+    def rshift(
+        self: "Parser[Output[A_monoid]]", p: "Parser[Output[B_monoid]]"
+    ) -> "Parser[Output[A_monoid | B_monoid]]":
+        return self >> p
+
     def __rshift__(
-        self: "Parser[typing.Sequence[A]]", p: "Parser[typing.Sequence[B]]"
-    ) -> "Parser[Sequence[A | B]]":
+        self: "Parser[Output[A_monoid]]", p: "Parser[Output[B_monoid]]"
+    ) -> "Parser[Output[A_monoid | B_monoid]]":
         """
         This applies parsers in sequence. If the first parser succeeds, the unparsed remainder
         gets handed off to the second parser. If either parser fails, the whole thing fails.
@@ -180,11 +176,11 @@ class Parser(MonadPlus[A_co]):
         The following arguments are required: second
         """
 
-        def f(p1: typing.Sequence[A]) -> Parser[Sequence[A | B]]:
-            def g(p2: typing.Sequence[B]) -> Parser[Sequence[A | B]]:
-                _p1 = p1 if isinstance(p1, Sequence) else Sequence(p1)
-                _p2 = p2 if isinstance(p2, Sequence) else Sequence(p2)
-                return Parser.return_(_p1 + _p2)
+        def f(p1: Output[A_monoid]) -> Parser[Output[A_monoid | B_monoid]]:
+            def g(p2: Output[B_monoid]) -> Parser[Output[A_monoid | B_monoid]]:
+                # _p1 = p1 if isinstance(p1, Sequence) else Sequence(p1)
+                # _p2 = p2 if isinstance(p2, Sequence) else Sequence(p2)
+                return Parser.return_(p1 + p2)
 
             return p >= g
 
@@ -197,8 +193,8 @@ class Parser(MonadPlus[A_co]):
         )
 
     def __xor__(
-        self: "Parser[Sequence[A]]", other: "Parser[Sequence[B]]"
-    ) -> "Parser[Sequence[A | B]]":
+        self: "Parser[Output[A_monoid]]", other: "Parser[Output[B_monoid]]"
+    ) -> "Parser[Output[A_monoid | B_monoid]]":
         """
         This is the same as `__or__`, but it succeeds only if one of the two parsers fails.
 
@@ -219,21 +215,22 @@ class Parser(MonadPlus[A_co]):
 
         return p.map_error(f)
 
-    def apply(self: "Parser[A]", f: Callable[[A], Result[B]]) -> "Parser[B]":  # type: ignore[misc]
+    def apply(self: "Parser[A_monoid]", f: Callable[[A_monoid], Result[B_monoid]]) -> "Parser[B_monoid]":  # type: ignore[misc]
+        # see https://github.com/python/mypy/issues/6178#issuecomment-1057111790
         """
         Takes the output of `parser` and applies `f` to it. Convert any errors that arise into `ArgumentError`.
 
         >>> p1 = flag("hello")
-        >>> p1.parse_args("--hello", return_dict=False)
-        [('hello', True)]
+        >>> p1.parse_args("--hello")
+        {'hello': True}
 
         This will double `p1`'s output:
-        >>> p2 = p1.apply(lambda kv: Result.return_(kv + kv))
-        >>> p2.parse_args("--hello", return_dict=False)
-        [('hello', True), ('hello', True)]
+        >>> p2 = p1.apply(lambda out: Result.return_(out + out))
+        >>> p2.parse_args("--hello")
+        {'hello': [True, True]}
         """
 
-        def g(a: A) -> Parser[B]:
+        def g(a: A_monoid) -> Parser[B_monoid]:
             try:
                 y = f(a)
             except Exception as e:
@@ -248,7 +245,7 @@ class Parser(MonadPlus[A_co]):
         p = self >= g
         return replace(p, usage=self.usage, helps=self.helps)
 
-    def bind(self, f: Callable[[A_co], Monad[B]]) -> "Parser[B]":
+    def bind(self, f: Callable[[A_co], Monad[B_monoid]]) -> "Parser[B_monoid]":  # type: ignore[override]
         """
         Returns a new parser that
 
@@ -267,8 +264,8 @@ class Parser(MonadPlus[A_co]):
 
         Now let's use the `matches` parser to write a function that takes the output of `p1` and fails unless
         the next argument is the same as the first:
-        >>> def f(kvs: Sequence[KeyValue[str]]) -> Parser[Sequence[KeyValue[str]]]:
-        ...     [kv] = kvs
+        >>> def f(out: Output[Sequence[KeyValue[Any]]]) -> Parser[Output[str]]:
+        ...     *_, kv = out.get
         ...     return matches(kv.value)
 
         >>> p = p1 >= f
@@ -278,26 +275,60 @@ class Parser(MonadPlus[A_co]):
         Expected 'a'. Got 'b'
         """
 
-        def h(parse: Parse[A_co]) -> Result[Parse[B]]:
+        def h(parse: Parse[A_co]) -> Result[Parse[B_monoid]]:
             y = f(parse.parsed)
             assert isinstance(y, Parser), y
             return y.parse(parse.unparsed)
 
-        def g(cs: Sequence[str]) -> Result[Parse[B]]:
+        def g(cs: Sequence[str]) -> Result[Parse[B_monoid]]:
             return self.parse(cs) >= h
 
         return Parser(g, usage=None, helps=self.helps)
 
     @classmethod
-    def empty(cls: Type["Parser[Sequence[A]]"]) -> "Parser[Sequence[A]]":
+    def done(
+        cls: Type["Parser[Output[A_monoid]]"], a: Optional[Type[A_monoid]] = None
+    ) -> Parser[Output[Any]]:
+        """
+        `done` succeds on the end of input and fails on everything else.
+        >>> Parser.done().parse_args()
+        {}
+        >>> Parser.done().parse_args("arg")
+        Unrecognized argument: arg
+
+        Without `done` the parser will not complain about leftover (unparsed) input:
+
+        >>> flag("verbose").parse_args("--verbose", "--quiet", allow_unparsed=True)
+        {'verbose': True}
+
+        When `allow_unparsed=False` (the default), `parse_args` adds `>> Parser.done()`
+        to the end of the parser:
+        """
+
+        def f(cs: Sequence[str]) -> Result[Parse[Output[Any]]]:
+            if cs:
+                c, *_ = cs
+                return Result(
+                    UnexpectedError(unexpected=c, usage=f"Unrecognized argument: {c}")
+                )
+            return Result(NonemptyList(Parse(parsed=Output.zero(a), unparsed=cs)))
+
+        return Parser(f, usage=None, helps={})
+
+    @classmethod
+    def empty(
+        cls: Type["Parser[Output[A_monoid]]"], a: Optional[Type[A_monoid]] = None
+    ) -> "Parser[Output[A_monoid]]":
         """
         Always returns {}, no matter the input. Mostly useful for use in `nonpositional`.
-        >>> empty().parse_args("any", "arguments", allow_unparsed=True)
+        >>> Parser.empty().parse_args("any", "arguments", allow_unparsed=True)
         {}
         """
-        return cls.return_(Sequence([]))
+        return cls.return_(Output.zero(a))
 
-    def fails(self: "Parser[Sequence[A]]") -> "Parser[Sequence[A]]":
+    def fails(
+        self: "Parser[Output[A_monoid]]", a: Optional[Type[A_monoid]] = None
+    ) -> "Parser[Output[A_monoid]]":
         """
         Succeeds only if self fails. Does not consume any input.
 
@@ -307,10 +338,10 @@ class Parser(MonadPlus[A_co]):
         Parser unexpectedly succeeded.
         """
 
-        def g(cs: Sequence[str]) -> Result[Parse[Sequence[A]]]:
+        def g(cs: Sequence[str]) -> Result[Parse[Output[A_monoid]]]:
             parse = self.parse(cs).get
             if isinstance(parse, Exception):
-                return Result.return_(Parse(Sequence([]), cs))
+                return Result.return_(Parse(Output.zero(a), cs))
             else:
                 return Result.zero(
                     error=SuccessError(
@@ -338,7 +369,9 @@ class Parser(MonadPlus[A_co]):
             if error.usage:
                 print(error.usage)
 
-    def ignore(self: "Parser[Sequence[A]]") -> "Parser[Sequence[A]]":
+    def ignore(
+        self: "Parser[Output[A_monoid]]", a: Optional[Type[A_monoid]] = None
+    ) -> "Parser[Output[A_monoid]]":
         """
         Ignores the output from a parser. This is useful when you expect
         to give arguments to the command line that some other utility will
@@ -355,37 +388,37 @@ class Parser(MonadPlus[A_co]):
         The following arguments are required: --hello
         """
 
-        def g(keep: Parse[Sequence[A]]) -> Result[Parse[Sequence[A]]]:
-            return Result(NonemptyList(Parse(Sequence([]), keep.unparsed)))
+        def g(keep: Parse[Output[A_monoid]]) -> Result[Parse[Output[A_monoid]]]:
+            return Result(NonemptyList(Parse(Output.zero(a), keep.unparsed)))
 
-        def f(cs: Sequence[str]) -> Result[Parse[Sequence[A]]]:
+        def f(cs: Sequence[str]) -> Result[Parse[Output[A_monoid]]]:
             return self.parse(cs) >= g
 
         return Parser(f, usage=None, helps={})
 
     def many(
-        self: "Parser[Sequence[A]]", max: Optional[int] = None
-    ) -> "Parser[Sequence[A]]":
+        self: "Parser[Output[A_monoid]]", max: Optional[int] = None
+    ) -> "Parser[Output[A_monoid]]":
         """
         Applies `self` zero or more times (like `*` in regexes).
 
         >>> from dollar_lambda import argument, flag
         >>> p = argument("as-many-as-you-like").many()
-        >>> p.parse_args(return_dict=False)
-        []
+        >>> p.parse_args()
+        {}
         >>> p = argument("as-many-as-you-like").many()
-        >>> p.parse_args("a", return_dict=False)
-        [('as-many-as-you-like', 'a')]
+        >>> p.parse_args("a")
+        {'as-many-as-you-like': 'a'}
         >>> p = argument("as-many-as-you-like").many()
-        >>> p.parse_args("a", "b", return_dict=False)  # return_dict=False allows duplicate keys
-        [('as-many-as-you-like', 'a'), ('as-many-as-you-like', 'b')]
+        >>> p.parse_args("a", "b")
+        {'as-many-as-you-like': ['a', 'b']}
 
         Note that if `self` contains `Parser.__or__`, the arguments can be
         heterogenous:
         >>> p = flag("verbose") | flag("quiet")
         >>> p = p.many()
-        >>> p.parse_args("--verbose", "--quiet", return_dict=False) # mix --verbose and --quiet
-        [('verbose', True), ('quiet', True)]
+        >>> p.parse_args("--verbose", "--quiet") # mix --verbose and --quiet
+        {'verbose': True, 'quiet': True}
         """
         if max == 0:
             p = self.empty()
@@ -397,8 +430,8 @@ class Parser(MonadPlus[A_co]):
         return replace(p, usage=f"[{self.usage} ...]")
 
     def many1(
-        self: "Parser[Sequence[A]]", max: Optional[int] = None
-    ) -> "Parser[Sequence[A]]":
+        self: "Parser[Output[A_monoid]]", max: Optional[int] = None
+    ) -> "Parser[Output[A_monoid]]":
         """
         Applies `self` one or more times (like `+` in regexes).
 
@@ -406,8 +439,8 @@ class Parser(MonadPlus[A_co]):
         >>> p = argument("1-or-more").many1()
         >>> p.parse_args("1")
         {'1-or-more': '1'}
-        >>> p.parse_args("1", "2", return_dict=False)  # return_dict=False allows duplicate keys
-        [('1-or-more', '1'), ('1-or-more', '2')]
+        >>> p.parse_args("1", "2")
+        {'1-or-more': ['1', '2']}
         >>> p.parse_args()
         usage: 1-OR-MORE [1-OR-MORE ...]
         The following arguments are required: 1-or-more
@@ -434,7 +467,7 @@ class Parser(MonadPlus[A_co]):
 
         return Parser(g, usage=None, helps=self.helps)
 
-    def optional(self: "Parser[Sequence[A]]") -> "Parser[Sequence[A]]":
+    def optional(self: "Parser[Output[A_monoid]]") -> "Parser[Output[A_monoid]]":
         """
         Allows arguments to be optional:
         >>> p1 = flag("optional")
@@ -456,12 +489,11 @@ class Parser(MonadPlus[A_co]):
         return self.f(cs)
 
     def parse_args(
-        self: "Parser[Sequence[KeyValue]]",
+        self: "Parser[Output]",
         *args: str,
         allow_unparsed: bool = False,
         check_help: bool = True,
-        return_dict: bool = True,
-    ) -> "typing.Sequence[KeyValueTuple] | Dict[str, Any]":
+    ) -> "Dict[str, Any]":
         """
         The main way the user extracts parsed results from the parser.
 
@@ -475,9 +507,6 @@ class Parser(MonadPlus[A_co]):
         check_help : bool
             Before running the parser, checks if the input string is `--help` or `-h`.
             If it is, returns the usage message.
-        return_dict : bool
-            Returns a sequence of tuples instead of dictionary, thereby allowing duplicate keys.
-            The tuples are `KeyValueTuple` namedtuples, with fields `key` and `value`.
 
         Examples
         --------
@@ -489,17 +518,15 @@ class Parser(MonadPlus[A_co]):
         """
         _args = args if args or TESTING else sys.argv[1:]
         if not allow_unparsed:
-            return (self >> done()).parse_args(
+            return (self >> Parser[Output].done()).parse_args(
                 *_args,
                 allow_unparsed=True,
                 check_help=check_help,
-                return_dict=return_dict,
             )
         if check_help:
             return self.wrap_help().parse_args(
                 *_args,
                 allow_unparsed=allow_unparsed,
-                return_dict=return_dict,
                 check_help=False,
             )
         result = self.parse(Sequence(list(_args))).get
@@ -510,10 +537,7 @@ class Parser(MonadPlus[A_co]):
             else:
                 exit()
 
-        kvs = result.head.parsed
-        if return_dict:
-            return {kv.key: kv.value for kv in kvs}
-        return [KeyValueTuple(**asdict(kv)) for kv in kvs]
+        return result.head.parsed.get.to_dict()
 
     @classmethod
     def return_(cls, a: A_co) -> "Parser[A_co]":  # type: ignore[misc]
@@ -523,9 +547,8 @@ class Parser(MonadPlus[A_co]):
         and always returns `a` as the result. For the most part, the user will not use
         this method unless building custom parsers.
 
-        >>> from dollar_lambda.key_value import KeyValue
-        >>> Parser.return_(([KeyValue("some-key", "some-value")])).parse_args()
-        {'some-key': 'some-value'}
+        >>> Parser.return_(Output.from_dict(some_key="some_value")).parse_args()
+        {'some_key': 'some_value'}
         """
 
         def f(cs: Sequence[str]) -> Result[Parse[A_co]]:
@@ -534,23 +557,23 @@ class Parser(MonadPlus[A_co]):
         return Parser(f, usage=None, helps={})
 
     def sat(
-        self: "Parser[A]",
-        predicate: Callable[[A], bool],
-        on_fail: Callable[[A], ArgumentError],
-    ) -> "Parser[A]":
+        self: "Parser[A_monoid]",
+        predicate: Callable[[A_monoid], bool],
+        on_fail: Callable[[A_monoid], ArgumentError],
+    ) -> "Parser[A_monoid]":
         """
         Applies `parser`, applies a predicate to the result and fails if this returns false.
 
         >>> p = option("x", type=int).many().sat(
-        ...     lambda kvs: sum([kv.value for kv in kvs]) > 0,
-        ...     lambda x: ArgumentError(f"The values in {list(x)} must sum to more than 0."),
+        ...     lambda out: sum(out.get.values()) > 0,
+        ...     lambda out: ArgumentError(f"The values in {list(out.get.values())} must sum to more than 0."),
         ... )
         >>> p.parse_args("-x", "-1", "-x", "1")  # fails
         usage: [-x X ...]
-        The values in [KeyValue(key='x', value=-1), KeyValue(key='x', value=1)] must sum to more than 0.
+        The values in [-1, 1] must sum to more than 0.
 
         >>> p.parse_args("-x", "-1", "-x", "2")  # succeeds
-        {'x': 2}
+        {'x': [-1, 2]}
 
         Parameters
         ----------
@@ -563,33 +586,33 @@ class Parser(MonadPlus[A_co]):
             Takes the output of `parser` as an argument.
         """
 
-        def f(x: A) -> Result[A]:
+        def f(x: A_monoid) -> Result[A_monoid]:
             return Result(NonemptyList(x) if predicate(x) else on_fail(x))
 
         return self.apply(f)
 
     def type(
-        self: "Parser[Sequence[KeyValue[str]]]", f: Callable[[str], Any]
-    ) -> "Parser[Sequence[KeyValue[Any]]]":
+        self: "Parser[Output[Sequence[KeyValue[str]]]]", f: Callable[[str], Any]
+    ) -> "Parser[Output[Sequence[KeyValue[str]]]]":
         """
         A wrapper around `apply` that simply applies `f` to the value of the most recently parsed input.
         >>> p1 = option("x") >> option("y")
         >>> p = p1.type(int)
         >>> p.parse_args("-x", "1", "-y", "2")  # converts "1" but not "2"
-        {'y': '2', 'x': 1}
+        {'x': '1', 'y': 2}
         """
 
-        def g(
-            kvs: Sequence[KeyValue[str]],
-        ) -> Result[Sequence[KeyValue[Any]]]:
-            head, *tail = kvs.get
+        def g(out: Output[Sequence[KeyValue[str]]]) -> Result[Output]:
+            d = out.get
+            if not d:
+                raise RuntimeError("Invoked type on a parser that returns no output.")
+            *tail, head = out.get
             try:
                 y = f(head.value)
             except Exception as e:
                 usage = f"argument {head.value}: raised exception {e}"
                 return Result(ArgumentError(usage))
-            head = replace(head, value=y)
-            return Result.return_(Sequence([*tail, head]))
+            return Result.return_(Output(Sequence([*tail, KeyValue(head.key, y)])))
 
         p = self.apply(g)
         return replace(p, usage=self.usage, helps=self.helps)
@@ -597,7 +620,9 @@ class Parser(MonadPlus[A_co]):
     def wrap_error(self, error: ArgumentError) -> "Parser[A_co]":
         return self.map_error(lambda _: error)
 
-    def wrap_help(self: "Parser[A]") -> "Parser[A]":
+    def wrap_help(
+        self: "Parser[A_monoid]", a: Optional[Type[A_monoid]] = None
+    ) -> "Parser[A_monoid]":
         """
         This checks for the `--help` or `-h` flag before applying `parser`.
         If either of the flags is present, returns the usage message for `parser`.
@@ -619,7 +644,7 @@ class Parser(MonadPlus[A_co]):
         >>> p.parse_args("subcommand2", "-h")
         usage: --option2 OPTION2
         """
-        _help_parser: Parser[Sequence[A]] = help_parser(self.usage, Sequence([]))
+        _help_parser: Parser[Output[A_monoid]] = help_parser(self.usage, Output.zero(a))
 
         p = _help_parser >= (lambda _: self)
         return replace(p, usage=self.usage, helps=self.helps)
@@ -644,7 +669,7 @@ class Parser(MonadPlus[A_co]):
         return Parser(lambda _: Result.zero(error=error), usage=None, helps={})
 
 
-def apply(f: Callable[[str], B], description: str) -> Parser[B]:
+def apply(f: Callable[[str], B_monoid], description: str) -> Parser[B_monoid]:
     """
     A shortcut for
     ```python
@@ -654,21 +679,22 @@ def apply(f: Callable[[str], B], description: str) -> Parser[B]:
     Here is an example of usage. First we define a simple `argument` parser:
 
     >>> p1 = argument("foo")
-    >>> p1.parse_args("bar", return_dict=False)
-    [('foo', 'bar')]
+    >>> p1.parse_args("bar")
+    {'foo': 'bar'}
 
     Here we use `f` to directly manipulate the binding generated by `item`:
-    >>> p2 = apply(lambda bar: [KeyValue(bar + "e", bar + "f")], description="baz")
-    >>> p2.parse_args("bar", return_dict=False)
-    [('bare', 'barf')]
+    >>> p2 = apply(lambda bar: Output.from_dict(**{bar + "e": bar + "f"}), description="baz")
+    >>> p2.parse_args("bar")
+    {'bare': 'barf'}
     """
 
-    def g(parsed: Sequence[KeyValue[str]]) -> Result[B]:
-        [kv] = parsed
+    def g(out: Output[Sequence[KeyValue[str]]]) -> Result[B_monoid]:
+        *_, (_, v) = map(astuple, out.get)
+        assert v is not None  # because item produces output
         try:
-            y = f(kv.value)
+            y = f(v)
         except Exception as e:
-            usage = f"argument {kv.value}: raised exception {e}"
+            usage = f"argument {v} raised exception {e}"
             return Result(ArgumentError(usage))
         return Result.return_(y)
 
@@ -677,7 +703,7 @@ def apply(f: Callable[[str], B], description: str) -> Parser[B]:
 
 def argument(
     dest: str, help: Optional[str] = None, type: Optional[Callable[[str], Any]] = None
-) -> Parser[Sequence[KeyValue[Any]]]:
+) -> Parser[Output]:
     """
     Parses a single word and binds it to `dest`.
     Useful for positional arguments.
@@ -701,6 +727,7 @@ def argument(
     """
     parser = item(dest)
     _type: Callable[[str], Any] = str if type is None else type  # type: ignore[assignment]
+    # Mypy doesn't know that types also have type Callable[[str], Any]
     if _type is not str:
         parser = parser.type(_type)
     helps = {dest: help} if help else {}
@@ -708,11 +735,11 @@ def argument(
     return parser
 
 
-def defaults(**kwargs: Any) -> Parser[Sequence[KeyValue[Any]]]:
+def defaults(**kwargs: A) -> Parser[Output[Sequence[KeyValue[A]]]]:
     """
     Useful for assigning default values to arguments.
-    It ignore the input and always returns `kwargs` converted into `Sequence[KeyValue]`.
-    `defaults` never fails.
+    It ignore the input and always returns `kwargs` converted into `CollisionDict`.
+    `defaults` never failsekpoi
 
     >>> defaults(a=1, b=2).parse_args()
     {'a': 1, 'b': 2}
@@ -727,46 +754,20 @@ def defaults(**kwargs: Any) -> Parser[Sequence[KeyValue[Any]]]:
     ...     ),
     ...     option("x", type=int, help="the base"),
     ...     option("y", type=int, help="the exponent"),
-    ... ) >> done()
+    ... )
 
     >>> p.parse_args("-x", "1", "-y", "2", "--verbose")
     {'x': 1, 'y': 2, 'verbose': True, 'quiet': False}
     """
-    p = Parser.return_(Sequence([KeyValue(k, v) for k, v in kwargs.items()]))
+    p = Parser[Output[A_monoid]].return_(
+        Output[Sequence[KeyValue[A]]].from_dict(**kwargs)
+    )
     return replace(p, usage=None)
 
 
-def done() -> Parser[Sequence[A]]:
-    """
-    `done` succeds on the end of input and fails on everything else.
-    >>> done().parse_args()
-    {}
-    >>> done().parse_args("arg")
-    Unrecognized argument: arg
-
-    Without `done` the parser will not complain about leftover (unparsed) input:
-
-    >>> flag("verbose").parse_args("--verbose", "--quiet", allow_unparsed=True)
-    {'verbose': True}
-
-    When `allow_unparsed=False` (the default), `parse_args` adds `>> done()`
-    to the end of the parser:
-    """
-
-    def f(cs: Sequence[str]) -> Result[Parse[Sequence[A]]]:
-        if cs:
-            c, *_ = cs
-            return Result(
-                UnexpectedError(unexpected=c, usage=f"Unrecognized argument: {c}")
-            )
-        return Result(NonemptyList(Parse(parsed=Sequence([]), unparsed=cs)))
-
-    return Parser(f, usage=None, helps={})
-
-
 def matches(
     s: str, peak: bool = False, regex: bool = True
-) -> Parser[Sequence[KeyValue[str]]]:
+) -> Parser[Output[Sequence[KeyValue[str]]]]:
     """
     Checks if the next word is `s`.
 
@@ -803,9 +804,9 @@ def matches(
 
     The first parser didn't consume the word and so "hello" got passed on to `equals("goodbye")`.
     But this would work:
-    >>> p = matches("hello", peak=True) >> matches("hello") >>matches("goodbye")
+    >>> p = matches("hello", peak=True) >> matches("hello") >> matches("goodbye")
     >>> p.parse_args("hello", "goodbye")
-    {'hello': 'hello', 'goodbye': 'goodbye'}
+    {'hello': ['hello', 'hello'], 'goodbye': 'goodbye'}
     """
 
     def predicate(_s: str) -> bool:
@@ -838,7 +839,7 @@ def flag(
     help: Optional[str] = None,
     short: bool = True,
     string: Optional[str] = None,
-) -> Parser[Sequence[KeyValue[bool]]]:
+) -> Parser[Output[Sequence[KeyValue[bool]]]]:
     """
     Binds a boolean value to a variable.
 
@@ -908,7 +909,7 @@ def flag(
     def f(
         cs: Sequence[str],
         s: str,
-    ) -> Result[Parse[Sequence[KeyValue[bool]]]]:
+    ) -> Result[Parse[Output[Sequence[KeyValue[bool]]]]]:
         parser = matches(s) >= (lambda _: defaults(**{dest: not default}))
         return parser.parse(cs)
 
@@ -924,10 +925,10 @@ def flag(
     return parser if default is None else parser | defaults(**{dest: default})
 
 
-def help_parser(usage: Optional[str], parsed: A) -> Parser[A]:
+def help_parser(usage: Optional[str], parsed: A_monoid) -> Parser[A_monoid]:
     def f(
         cs: Sequence[str],
-    ) -> Result[Parse[A]]:
+    ) -> Result[Parse[A_monoid]]:
         result = (matches("--help", peak=True) | matches("-h", peak=True)).parse(cs)
         if isinstance(result.get, ArgumentError):
             return Result.return_(Parse(parsed=parsed, unparsed=cs))
@@ -939,7 +940,7 @@ def help_parser(usage: Optional[str], parsed: A) -> Parser[A]:
 def item(
     name: str,
     help_name: Optional[str] = None,
-) -> Parser[Sequence[KeyValue[str]]]:
+) -> Parser[Output[Sequence[KeyValue[str]]]]:
     """
     Parses a single word and binds it to `dest`.
     One of the lowest level building blocks for parsers.
@@ -960,15 +961,15 @@ def item(
     The following arguments are required: Your first name
     """
 
-    def f(
-        cs: Sequence[str],
-    ) -> Result[Parse[Sequence[KeyValue[str]]]]:
+    def f(cs: Sequence[str]) -> Result[Parse[Output[Sequence[KeyValue[str]]]]]:
         if cs:
             head, *tail = cs
             return Result(
                 NonemptyList(
                     Parse(
-                        parsed=Sequence([KeyValue(name, head)]),
+                        parsed=Output[Sequence[KeyValue[str]]].from_dict(
+                            **{name: head}
+                        ),
                         unparsed=Sequence(tail),
                     )
                 )
@@ -984,8 +985,9 @@ def item(
 
 
 def nonpositional(
-    *parsers: "Parser[Sequence[A]]", repeated: Optional[Parser[Sequence[A]]] = None
-) -> "Parser[Sequence[A]]":
+    *parsers: "Parser[Output[A_monoid]]",
+    repeated: Optional[Parser[Output[A_monoid]]] = None,
+) -> "Parser[Output[A_monoid]]":
     """
     `nonpositional` takes a sequence of parsers as arguments and attempts all permutations of them,
     returning the first permutations that is successful:
@@ -1004,30 +1006,30 @@ def nonpositional(
     Examples
     --------
     >>> p = nonpositional(repeated=flag("x"))
-    >>> p.parse_args(return_dict=False)
-    []
-    >>> p.parse_args("-x", return_dict=False)
-    [('x', True)]
-    >>> p.parse_args("-x", "-x", return_dict=False)
-    [('x', True), ('x', True)]
+    >>> p.parse_args()
+    {}
+    >>> p.parse_args("-x")
+    {'x': True}
+    >>> p.parse_args("-x", "-x")
+    {'x': [True, True]}
 
     >>> p = nonpositional(flag("y"), repeated=flag("x"))
-    >>> p.parse_args("-y", return_dict=False)
-    [('y', True)]
-    >>> p.parse_args("-y", "-x", return_dict=False)
-    [('y', True), ('x', True)]
-    >>> p.parse_args("-x", "-y", return_dict=False)
-    [('x', True), ('y', True)]
-    >>> p.parse_args("-y", "-x", "-x", return_dict=False)
-    [('y', True), ('x', True), ('x', True)]
-    >>> p.parse_args("-x", "-y", "-x", return_dict=False)
-    [('x', True), ('y', True), ('x', True)]
-    >>> p.parse_args("-x", "-x", "-y", return_dict=False)
-    [('x', True), ('x', True), ('y', True)]
+    >>> p.parse_args("-y")
+    {'y': True}
+    >>> p.parse_args("-y", "-x")
+    {'y': True, 'x': True}
+    >>> p.parse_args("-x", "-y")
+    {'x': True, 'y': True}
+    >>> p.parse_args("-y", "-x", "-x")
+    {'y': True, 'x': [True, True]}
+    >>> p.parse_args("-x", "-y", "-x")
+    {'x': [True, True], 'y': True}
+    >>> p.parse_args("-x", "-x", "-y")
+    {'x': [True, True], 'y': True}
 
     >>> p = nonpositional(flag("y"), repeated=(flag("x") | flag("z")).ignore())
-    >>> p.parse_args("-x", "-y", "-z", return_dict=False)
-    [('y', True)]
+    >>> p.parse_args("-x", "-y", "-z")
+    {'y': True}
     """
     sep = " " if len(parsers) <= 3 else "\n"
     _parsers = [*parsers] if repeated is None else [*parsers, repeated]
@@ -1044,9 +1046,9 @@ def nonpositional(
 
     if not parsers:
         if repeat_parser is None:
-            return empty()
+            return Parser[Output[A_monoid]].empty()
         else:
-            return repeat_parser | empty()
+            return repeat_parser | Parser[Output[A_monoid]].empty()
 
     def get_alternatives():
         if repeat_parser is not None:
@@ -1069,7 +1071,7 @@ def option(
     help: Optional[str] = None,
     short: bool = True,
     type: Callable[[str], Any] = str,
-) -> Parser[Sequence[KeyValue[str]]]:
+) -> Parser[Output[Sequence[KeyValue[Any]]]]:
     """
     Parses two words, binding the second to the first.
 
@@ -1142,7 +1144,7 @@ def option(
 
     def f(
         cs: Sequence[str],
-    ) -> Result[Parse[Sequence[KeyValue[str]]]]:
+    ) -> Result[Parse[Output[Sequence[KeyValue[str]]]]]:
         parser = matches(_flag) >= (lambda _: item(dest, help_name=dest.upper()))
         return parser.parse(cs)
 
@@ -1160,16 +1162,18 @@ def option(
 def peak(
     name: str,
     description: Optional[str] = None,
-) -> Parser[Sequence[KeyValue[str]]]:
+) -> Parser[Output[Sequence[KeyValue[str]]]]:
     def f(
         cs: Sequence[str],
-    ) -> Result[Parse[Sequence[KeyValue[str]]]]:
+    ) -> Result[Parse[Output[Sequence[KeyValue[str]]]]]:
         if cs:
             head, *_ = cs
             return Result(
                 NonemptyList(
                     Parse(
-                        parsed=Sequence([KeyValue(name, head)]),
+                        parsed=Output[Sequence[KeyValue[str]]].from_dict(
+                            **{name: head}
+                        ),
                         unparsed=Sequence(cs),
                     )
                 )
@@ -1188,7 +1192,7 @@ def sat(
     predicate: Callable[[str], bool],
     on_fail: Callable[[str], ArgumentError],
     name: str,
-) -> Parser[Sequence[KeyValue[str]]]:
+) -> Parser[Output[Sequence[KeyValue[str]]]]:
     """
     A wrapper around `Parser.sat` that uses `item` to parse the argument and just applies `predicate` to the value output by `item`.
 
@@ -1210,13 +1214,13 @@ def sat(
         The value to bind the result to.
     """
 
-    def _predicate(parsed: Sequence[KeyValue[str]]) -> bool:
-        [kv] = parsed
-        return predicate(kv.value)
+    def _predicate(out: Output[Sequence[KeyValue[str]]]) -> bool:
+        *_, (_, v) = map(astuple, out.get)
+        return predicate(v)
 
-    def _on_fail(parsed: Sequence[KeyValue[str]]) -> ArgumentError:
-        [kv] = parsed
-        return on_fail(kv.value)
+    def _on_fail(out: Output[Sequence[KeyValue[str]]]) -> ArgumentError:
+        *_, (_, v) = map(astuple, out.get)
+        return on_fail(v)
 
     return item(name).sat(_predicate, _on_fail)
 
@@ -1225,13 +1229,13 @@ def sat_peak(
     predicate: Callable[[str], bool],
     on_fail: Callable[[str], ArgumentError],
     name: str,
-) -> Parser[Sequence[KeyValue[str]]]:
-    def _predicate(parsed: Sequence[KeyValue[str]]) -> bool:
-        [kv] = parsed
-        return predicate(kv.value)
+) -> Parser[Output[Sequence[KeyValue[str]]]]:
+    def _predicate(out: Output[Sequence[KeyValue[str]]]) -> bool:
+        *_, (_, v) = map(astuple, out.get)
+        return predicate(v)
 
-    def _on_fail(parsed: Sequence[KeyValue[str]]) -> ArgumentError:
-        [kv] = parsed
-        return on_fail(kv.value)
+    def _on_fail(out: Output[Sequence[KeyValue[str]]]) -> ArgumentError:
+        *_, (_, v) = map(astuple, out.get)
+        return on_fail(v)
 
     return peak(name).sat(_predicate, _on_fail)
